{- 

CS457/557 Functional Languages Spring 2018 

Lecture 10: Monads

Andrew Tolmach

-}

import System.IO(isEOF)


{-

Reviewing IO Actions
--------------------

- Recall properties of special type of IO actions.

- Basic operations have "side effects," e.g.

  getChar :: IO Char
  putChar :: Char -> IO ()
  isEOF   :: IO Bool

- Operations are combined into sequences using "do":

-}

echo :: IO ()
echo = do b <- isEOF
          if not b then
           do x <- getChar
              putChar x
              echo
          else return ()
         
{-

Operations don't actually happend except at "top level"
where we IMPLICITLY perform an operation with type

 runIO :: IO a -> a  -- actully do the IO

(But this operator doesn't really exist.)

The Real Primitives
-------------------

The "do" notation is just syntactic sugar for as sequence
of applications of a particular primitive operator written
>>= and pronounced "bind".

The fundamental primitives are:

 (>>=)  :: IO a -> (a -> IO b) -> IO b
 return :: a -> IO a

These can be used just like any other functions.

 p >>= q 

is an IO action that runs p, pipes the output into q,
and runs the resulting action...

The "do" notation expands to >>=

For example:

   do x1 <- p1
      ...
      xn <- pn
      q


is equivalent to

   p1 >>= (\x1 ->
   ...
   pn >>= (\xn ->
   q) ...)

We can also drop parentheses and write:

   p1 >>= \x1 ->
   ...
   pn >>= \xn ->
   q

Still, the "do" notation is usually nicer.

A special case of bind that ignores the result of the first 
computation::

 (>>) :: IO a -> IO b -> IO b
 p >> q = p >>= (\_ -> q)

The output of p is ignored by q.

The full "do" translation looks like this:

  do {e} ≡ e
  do {e; es} ≡ e >> do {es}
  do {x <- e; es} ≡ e >>= (\x ->  do {es})
  do {let ds; es} ≡ let ds in do {es}

-}


   
