data Exp =
     Plus Exp Exp
   | Minus Exp Exp
   | Times Exp Exp
   | Div Exp Exp
   | Const Int

----------------------


{- Suppose we want to improve this behavior by trapping
   division by zero.  Now 'eval' returns either a numeric
   value (Ok v) or an Error value describing what went
   wrong during evaluation. 
 -}

data Exn a = Ok a | Error String      
 deriving Show

eval :: Exp -> Exn Int
eval (Plus e1 e2)  =
     case eval e1 of
       Ok v1 ->
          case eval e2 of
            Ok v2 -> Ok (v1+v2)
            Error s -> Error s
       Error s -> Error s
eval (Minus e1 e2) = 
     case eval e1 of
       Ok v1 ->
          case eval e2 of
            Ok v2 -> Ok (v1-v2)
            Error s -> Error s
       Error s -> Error s
eval (Times e1 e2) = 
     case eval e1 of
       Ok v1 ->
          case eval e2 of
            Ok v2 -> Ok (v1*v2)
            Error s -> Error s
       Error s -> Error s
eval (Div e1 e2)   =
     case eval e1 of
       Ok v1 ->
          case eval e2 of
            Ok v2 | v2 == 0 -> Error "divby0"
                  | otherwise -> Ok(v1 `div` v2)
            Error s -> Error s
       Error s -> Error s
eval (Const i)     = Ok i

example1 = eval (Div (Plus (Const 5) (Const 3))
                     (Const 2))

example2 = eval (Minus (Const 4) (Div (Const 2) (Const 0)))


{- This solution exposes a lot of ugly plumbing!

   (Of course, we could abstract over the different arithmetic operators to
   improve code reuse, but that wouldn't help if we had a wider range
   of interesting statement forms.)
   
   Notice that whenever an expression evalutes to Error,
   that Error propagates up to the final result.


 -}

