data Exp =
     Plus Exp Exp
   | Minus Exp Exp
   | Times Exp Exp
   | Div Exp Exp
   | Const Int


data Exn a = Ok a | Error String      
 deriving Show

-----------

{-

 We can abstract this propagation behavior into a 
 higher-order function:

-}

andthen :: Exn a -> (a -> Exn b) -> Exn b
e `andthen` k =
  case e of
    Ok x    -> k x
    Error s -> Error s

eval :: Exp -> Exn Int
eval (Plus e1 e2)  =
     eval e1 `andthen` (\v1 ->
     eval e2 `andthen` (\v2 ->      
     Ok (v1+v2)))
eval (Minus e1 e2) = 
     eval e1 `andthen` (\v1 ->
     eval e2 `andthen` (\v2 ->      
     Ok (v1-v2)))
eval (Times e1 e2) = 
     eval e1 `andthen` (\v1 ->
     eval e2 `andthen` (\v2 ->      
     Ok (v1*v2)))
eval (Div e1 e2)   =
     eval e1 `andthen` (\v1 ->
     eval e2 `andthen` (\v2 ->      
     if v2 == 0 then
       Error "divby0"
     else Ok (v1 `div` v2)))
eval (Const i)     = Ok i

example1 = eval (Div (Plus (Const 5) (Const 3))
                     (Const 2))

example2 = eval (Minus (Const 4) (Div (Const 2) (Const 0)))


{- 

Exn and IO are both MONADS
--------------------------


Compare the types of these functions:

  andthen :: Exn a -> (a -> Exn b) -> Exn b
  Ok      :: a -> Exn a


  (>>=) :: IO a -> (a -> IO b) -> IO b
  return :: a -> IO a


The similarities are not accidental!

IO, Exn, and many other type constructors are instances of
a more general structure called a MONAD.

Monads are useful for describing many kinds of
COMPUTATIONAL EFFECTS where there is a concept of
SEQUENCING (captured by >>=).

-}

