data Exp =
     Plus Exp Exp
   | Minus Exp Exp
   | Times Exp Exp
   | Div Exp Exp
   | Const Int

-------------------


{- Now let's revert to our original interpreter, and
   extend it in a new way, by accumulating a trace of
   operations in the order that they occur.

   In an imperative language, it would be natural to
   build up the trace in a global variable that is
   appended to by every eval stop.

   But in a functional program, we seemingly must 
   write something like this:
-}

eval :: Exp -> (String,Int)
eval (Plus e1 e2)  =
     let (s1,v1) = eval e1
         (s2,v2) = eval e2
     in (s1 ++ s2 ++ " +", v1 + v2)
eval (Minus e1 e2)  =
     let (s1,v1) = eval e1
         (s2,v2) = eval e2
     in (s1 ++ s2 ++ " -", v1 - v2)
eval (Times e1 e2)  =
     let (s1,v1) = eval e1
         (s2,v2) = eval e2
     in (s1 ++ s2 ++ " *", v1 * v2)
eval (Div e1 e2)  =
     let (s1,v1) = eval e1
         (s2,v2) = eval e2
     in (s1 ++ s2 ++ " /", v1 `div` v2)
eval (Const i)     = (" " ++ show i,i)

example1 = eval (Div (Plus (Const 5) (Const 3))
                     (Const 2))


{- Again, there is a lot of unpleasant plumbing.
   Can we avoid that?

-}

