import Control.Monad

{- 

Simulating the IO Monad
-----------------------

The IO monad is "built in" to Haskell, i.e.
it cannot be implemented within the langauge itself:

- Special primitives are needed to actually perform
the IO actions and to sequence them.

- The IO type is abstract (it has no constructors).

But we can SIMULATE the behavior of IO actions (on
a single IO channel) using a state monad:

-}

newtype IOX t = IOX (Input -> (t, Input, Output))
type Input = String
type Output = String

{- 

Each IOX function takes the available input as argument,
performs an action that consumes some of that input, 
and returns;

- the result of the action (of type t)
- the remaining input
- any output produced by the action

-}


instance Monad IOX where
 (IOX m) >>= k =
   IOX (\input -> 
     let (t,input',output) = m input
         IOX m' = k t
         (t',input'',output') = m' input'
     in (t',input'',output ++ output'))
 return x = IOX (\input -> (x,input,""))
runIOX :: IOX () -> Input -> Output
runIOX (IOX m) input = output 
  where (_,_,output) = m input

instance Functor IOX where
  fmap = liftM
instance Applicative IOX where
  pure = return 
  (<*>) = ap


getChar' :: IOX Char
getChar' = IOX (\(i:is) -> (i,is,""))
putChar' :: Char -> IOX ()
putChar' c = IOX (\is -> ((),is,[c]))
isEOF' :: IOX Bool
isEOF' = IOX(\input ->
              (null input,input,""))

copy :: IOX ()
copy = do b <- isEOF'
          if b then return ()
          else do c <- getChar'
                  putChar' c
                  copy


example = runIOX copy "abc"
